import .imp_syntax

namespace imp

-- Stores

def store.empty : store := []

def update (x : string) (v : val) : store → store
| [] := [(x, v)]
| ((x', v') :: rest) :=
  if x = x'
  then (x, v) :: rest
  else (x', v') :: (update rest).

def updates : list string → list val →  store → option store
| [] [] s := some s
| (x :: xs') (v :: vs') s := updates xs' vs' (update x v s)
| _ _ _ := none

def store.lookup (x : string) : store → option val
| [] := none
| ((x', v) :: rest) :=
  if x = x'
  then some v
  else store.lookup rest

-- Common
def heap.size : heap → ℤ
| [] := 0
| (_ :: rest) := 1 + heap.size rest.

def is_nop (p : stmt) : decidable (p = stmt.nop) :=
begin
  refine (if H : p = stmt.nop then _ else _),
  apply decidable.is_true,
  assumption,
  apply decidable.is_false,
  assumption,
end

open stmt

def stmt.size : stmt → nat
| nop := 1
| (set _ _) := 1
| (ifelse _ p1 p2) := nat.succ (stmt.size p1 + stmt.size p2)
| (while _ p1) := nat.succ (stmt.size p1)
| (seq p1 p2) := nat.succ (stmt.size p1 + stmt.size p2)

end imp
