import .imp_eval
import .imp_syntax
import .imp_step
import .tactics

namespace imp

lemma eval_unop_det :
  forall op v v1 v2,
    eval_unop op v v1 ->
    eval_unop op v v2 ->
    v1 = v2 :=
begin
  intros,
  cases a; cases a_1; simp,
end

lemma eval_binop_det :
  forall op vL vR v1 v2,
    eval_binop op vL vR v1 ->
    eval_binop op vL vR v2 ->
    v1 = v2 :=
begin
  intros,
  cases a; cases a_1; simp
end

lemma eval_e_det :
  forall s e v1 v2,
    eval_e s e v1 →
    eval_e s e v2 →
    v1 = v2 :=
begin
  intros s e v1 v2 H,
  revert v2,
  induction H; intros,
  case eval_e.val {
    cases a,
    reflexivity
  },
  case eval_e.var {
    cases a_1,
    simp * at *,
    cases a_2,
    reflexivity
  },
  case eval_e.unary_op {
    cases a_2,
    have x := ih_1 v_1 a_3,
    rw x at a_1,
    apply eval_unop_det ; assumption
  },
  case eval_e.binary_op {
    cases a_3,
    apply eval_binop_det,
    apply a_2,
    specialize (ih_1 _ a_4),
    specialize (ih_2 _ a_5),
    simp * at *,
  }
end

lemma nostep_nop :
  forall s s' p',
    ¬ step
      s stmt.nop
      s' p' :=
begin
  delta not, intros, cases a,
end

lemma stmt_ifelse_neq_self :
  forall c b1 b2,
    let t := stmt.ifelse c b1 b2
    in b1 ≠ t ∧ b2 ≠ t :=
begin
  dsimp, intros,
  split,
  induction b1 ; try {cc },
  delta not ; intros,
  delta not at ih_1,
  apply ih_1,
  cc,
  delta not ; intros,
  induction b2 ; try { cc },
end

lemma stmt_seq_neq_self :
  forall t1 t2,
    stmt.seq t1 t2 ≠ t1 ∧
    stmt.seq t1 t2 ≠ t2 :=
begin
  intros, split,
  induction t1 ; try { cc },
  dsimp * at *, delta not at *,
  intros, apply ih_1,
  cc,
  induction t2 ; try { cc },
  dsimp * at *, delta not at *,
  intros, apply ih_2,
  cc,
end

lemma nostep_self' :
  forall s1 p1 s2 p2,
    step
      s1 p1
      s2 p2 ->
    p1 ≠ p2 :=
begin
  intros,
  induction a ; try { cc },
  have xs := (stmt_ifelse_neq_self e p1_1 p2_1),
  dsimp at xs, cc,
  have xs := (stmt_ifelse_neq_self e p1_1 p2_1),
  dsimp at xs, cc,
  have xs := (stmt_seq_neq_self stmt.nop p2_1),
  cc,
  have xs := (stmt_seq_neq_self stmt.nop p2_1),
  delta ne, delta not, intros,
  cases a_1, cc
end

lemma nostep_self :
  forall s1 p1 s2,
    ¬ step
      s1 p1
      s2 p1 :=
begin
  delta not,
  intros,
  apply nostep_self'; [assumption, reflexivity]
end

lemma step_star_1 :
  forall s1 p1 s2 p2,
    step
      s1 p1
      s2 p2 ->
    step_star
      s1 p1
      s2 p2 :=
begin
  intros,
  constructor,
  assumption,
  constructor,
end

lemma step_star_r :
  forall s1 p1 s2 p2 s3 p3,
    step_star
      s1 p1
      s2 p2 ->
    step
      s2 p2
      s3 p3 ->
    step_star
      s1 p1
      s3 p3 :=
begin
  intros,
  revert a_1 s3 p3,
  induction a; intros,
  case step_star.refl {
    intros,
    constructor,
    assumption,
    constructor
  },
  case step_star.left {
    constructor,
    assumption,
    apply ih_1,
    assumption,
  }
end

lemma step_star_app :
  forall s1 p1 s2 p2 s3 p3,
    step_star
      s1 p1
      s2 p2 ->
    step_star
      s2 p2
      s3 p3 ->
    step_star
      s1 p1
      s3 p3 :=
begin
  intros,
  revert a_1,
  revert p3,
  revert s3,
  induction a; intros,
  { assumption },
  { constructor,
    assumption,
    apply ih_1 ; assumption
  }
end

lemma step_star_seq :
  forall s1 p1 s2 p2 pB,
    step_star
      s1 p1
      s2 p2 ->
    step_star
      s1 (stmt.seq p1 pB)
      s2 (stmt.seq p2 pB) :=
begin
  intros,
  revert pB,
  induction a ; intros,
  case step_star.refl { constructor },
  case step_star.left {
    constructor,
    constructor,
    assumption,
    apply ih_1,
  }
end

lemma eval_step_star :
  forall s p s',
    eval_s
      s p s' ->
    step_star
      s p
      s' stmt.nop :=
begin
  intros ; induction a,
  case eval_s.nop { constructor },
  case eval_s.set {
    constructor,
    constructor,
    assumption,
    constructor,
  },
  case eval_s.ifelse_t {
    constructor,
    apply step.ifelse_t,
    assumption,
    assumption,
  },
  case eval_s.ifelse_f {
    constructor,
    apply step.ifelse_f,
    assumption,
    assumption
  },
  case eval_s.while_t {
    apply step_star.left,
    constructor ; assumption,
    apply step_star_app,
    apply step_star_seq ; assumption,
    apply step_star.left,
    constructor,
    assumption,
  },
  case eval_s.while_f {
    apply step_star_app,
    apply step_star.left,
    apply step.while_f,
    assumption,
    constructor,
    constructor,
  },
  case eval_s.seq {
    apply step_star_app,
    apply step_star_seq,
    assumption,
    constructor,
    constructor,
    assumption
  }
end

lemma step_eval_eval :
  forall s1 p1 s2 p2 s3,
    step
      s1 p1
      s2 p2 ->
    eval_s
      s2 p2 s3 ->
    eval_s
      s1 p1 s3 :=
begin
  intros,
  revert s3,
  induction a ; intros,
  case step.set {
      cases a_1,
      constructor,
      assumption,
  },
  case step.ifelse_t {
    constructor ; assumption,
  },
  case step.ifelse_f {
    apply eval_s.ifelse_f ; assumption,
  },
  case step.while_t {
    cases a_1,
    constructor ; assumption,
  },
  case step.while_f {
    cases a_1,
    apply eval_s.while_f ; assumption
  },
  case step.seq_nop {
      constructor,
      constructor,
      assumption,
   },
  case step.seq {
    cases a_1,
    constructor,
    apply ih_1,
    assumption,
    assumption,
  }
end

lemma step_star_eval :
  forall s1 p1 s2,
    step_star
      s1 p1
      s2 stmt.nop ->
    eval_s
      s1 p1 s2 :=
begin
  intros,
  induction_on a,
  { subst eq, constructor },
  { subst eq, apply step_eval_eval; try { assumption },
    apply ih_1, reflexivity
  }
end

end imp
