namespace imp

inductive val : Type
| bool : bool → val
| int  : ℤ → val
| str  : string → val

instance val.decidable_eq : decidable_eq val :=
by tactic.mk_dec_eq_instance

inductive unary_op : Type
| neg
| not

instance unary_op.decidable_eq : decidable_eq unary_op :=
by tactic.mk_dec_eq_instance

inductive binary_op : Type
| add
| sub
| mul
| div
| mod
| eq
| lt
| le
| conj
| disj

instance binary_op.decidable_eq : decidable_eq binary_op :=
by tactic.mk_dec_eq_instance

inductive expr : Type
| val : val -> expr
| var : string -> expr
| unary_op : unary_op -> expr -> expr
| binary_op : binary_op -> expr -> expr -> expr

instance expr.decidable_eq : decidable_eq expr :=
by tactic.mk_dec_eq_instance

@[reducible] def store : Type :=
list (string × val)

instance store.decidable_eq : decidable_eq store :=
by tactic.mk_dec_eq_instance

@[reducible] def heap : Type :=
list val

instance heap.decidable_eq : decidable_eq heap :=
by tactic.mk_dec_eq_instance

inductive stmt : Type
| nop : stmt
| set : string -> expr -> stmt
| ifelse : expr -> stmt -> stmt -> stmt
| while : expr -> stmt -> stmt
| seq : stmt -> stmt -> stmt

instance stmt.decidable_eq : decidable_eq stmt :=
by tactic.mk_dec_eq_instance

end imp
