import .imp_syntax
import .imp_common
import .imp_eval
import .imp_step
import .imp_semantics_facts

namespace imp

 def always_diverges (p : stmt) : Prop :=
  forall (s s' : store) (p' : stmt),
    step_star s p s' p' ->
    p' ≠ stmt.nop

def while_true_nop : stmt :=
  stmt.while (expr.val $ val.bool tt) stmt.nop.

lemma while_true_nop_diverges :
    always_diverges while_true_nop :=
begin
  unfold always_diverges while_true_nop ; intros,
  induction_on a,
  { rw <- eq, cc }
  { /- check IH, not inductive! -/ }
end

lemma while_t_sstar_cases' :
  forall s s' p',
    step_star
      s while_true_nop
      s' p' ->
    p' = stmt.seq stmt.nop while_true_nop ∨
    p' = while_true_nop :=
begin
  unfold while_true_nop ; intros,
  induction a,
  { right, reflexivity },
  { /- check IH, still not inductive! -/ }
end

def while_true_nop' : stmt :=
  stmt.seq stmt.nop (stmt.while (expr.val (val.bool tt)) stmt.nop)

lemma while_t_sstar_cases :
  forall s p s' p',
    step_star
      s p
      s' p' ->
    (p = while_true_nop \/ p = while_true_nop') ->
    p' = while_true_nop \/ p' = while_true_nop' :=
begin
  intros s p s' p' H,
  unfold while_true_nop while_true_nop',
  induction H ; intros,
  { cases a,
    left, assumption,
    right, assumption },
  { apply ih_1,
    clear ih_1,
    cases a_2,
    { subst a_2,
      cases a,
      clear a, clear a_1, clear a_2,
      right, reflexivity,
      cases a_2
    },
    { subst a_2, cases a,
      { left, reflexivity },
      { cases a_2, } }
  }
end

lemma while_t_diverges :
    always_diverges while_true_nop :=
begin
  unfold always_diverges while_true_nop ; intros,
  have wtn_or_wtn' := (while_t_sstar_cases _ _ _ _ a _),
  cases wtn_or_wtn',
  subst a_1, contradiction,
  subst a_1, contradiction,
  left, reflexivity
end

end imp
